"""Two dimensional convolution based fast wavelet transforms."""
#
# Created on Thu Jun 12 2020
# Copyright (c) 2020 Moritz Wolter
#
from typing import List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import pywt

from .conv_fwt import _get_filter_arrays
from .utils import _as_wavelet


def wavedec2(
    data: jnp.ndarray,
    wavelet: pywt.Wavelet,
    level: Optional[int] = None,
    mode: str = "reflect",
    precision: str = "highest",
) -> List[Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]]:
    """Compute the two dimensional wavelet analysis transform on the last two dimensions of the input data array.

    Args:
        data (jnp.ndarray): Jax array containing the data to be transformed. Assumed shape:
                         [batch size, hight, width].
        wavelet (pywt.Wavelet): A namedtouple containing the filters for the transformation.
        level (int): The max level to be used, if not set as many levels as possible
                               will be used. Defaults to None.
        mode (str): The desired padding mode. Choose reflect, symmetric or zero.
            Defaults to reflect.
        precision (str): The desired precision, choose "fastest", "high" or "highest".
            Defaults to "highest".

    Returns:
        list: The wavelet coefficients of shape [batch, height, width] in a nested list.
            The coefficients are in pywt order. That is:
            [cAn, (cHn, cVn, cDn), … (cH1, cV1, cD1)].
            A denotes approximation, H horizontal, V vertical
            and D diagonal coefficients.

    Raises:
        ValueError: If the dimensionality of the input data array is unsupported.

    Examples:
        >>> import pywt, scipy.datasets
        >>> import jaxwt as jwt
        >>> import jax.numpy as jnp
        >>> face = jnp.transpose(scipy.datasets.face(), [2, 0, 1])
        >>> face = face.astype(jnp.float64)
        >>> jwt.wavedec2(face, pywt.Wavelet("haar"), level=2)
    """
    wavelet = _as_wavelet(wavelet)

    if len(data.shape) == 2:
        data = jnp.expand_dims(data, (0, 1))
    elif len(data.shape) == 3:
        data = jnp.expand_dims(data, 1)
    elif len(data.shape) == 4:
        raise ValueError(
            "Wavedec2 does not support four input dimensions. \
             Optionally-batched two dimensional inputs work."
        )
    elif len(data.shape) == 1:
        raise ValueError("Wavedec2 needs more than one input dimension to work.")

    # data = jnp.expand_dims(data, 1)
    dec_lo, dec_hi, _, _ = _get_filter_arrays(wavelet, flip=True, dtype=data.dtype)
    dec_filt = _construct_2d_filt(lo=dec_lo, hi=dec_hi)

    if mode == "zero":
        # translate pywt to numpy.
        mode = "constant"

    if level is None:
        level = pywt.dwtn_max_level(
            [data.shape[-1], data.shape[-2]], pywt.Wavelet("MyWavelet", wavelet)
        )

    result_lst: List[
        Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]
    ] = []
    res_ll = data
    for _ in range(level):
        res_ll = _fwt_pad2d(res_ll, len(wavelet), mode=mode)
        res = jax.lax.conv_general_dilated(
            lhs=res_ll,  # lhs = NCHw image tensor
            rhs=dec_filt,  # rhs = OIHw conv kernel tensor
            padding="VALID",
            window_strides=[2, 2],
            dimension_numbers=("NCHW", "OIHW", "NCHW"),
            precision=jax.lax.Precision(precision),
        )
        res_ll, res_lh, res_hl, res_hh = jnp.split(res, 4, 1)
        result_lst.append((res_lh.squeeze(1), res_hl.squeeze(1), res_hh.squeeze(1)))
    result_lst.append(res_ll.squeeze(1))
    result_lst.reverse()
    return result_lst


def waverec2(
    coeffs: List[Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]],
    wavelet: pywt.Wavelet,
    precision: str = "highest",
) -> jnp.ndarray:
    """Compute a two dimensional synthesis wavelet transfrom.

       Use it to reconstruct the original input image from the wavelet coefficients.

    Args:
        coeffs (list): The input coefficients, typically the output of wavedec2.
        wavelet (pywt.Wavelet): The named tuple contining the filters used to compute the analysis transform.
        precision (str): The desired precision, choose "fastest", "high" or "highest".
            Defaults to "highest".

    Returns:
        jnp.array: Reconstruction of the original input data array of shape [batch, height, width].

    Raises:
        ValueError: If `coeffs` is not in the shape as it is returned from `wavedec2`.

    Example:
        >>> import pywt, scipy.datasets
        >>> import jaxwt as jwt
        >>> import jax.numpy as jnp
        >>> face = jnp.transpose(scipy.datasets.face(), [2, 0, 1])
        >>> face = face.astype(jnp.float64)
        >>> transformed = jwt.wavedec2(face, pywt.Wavelet("haar"))
        >>> jwt.waverec2(transformed, pywt.Wavelet("haar"))


    """
    wavelet = _as_wavelet(wavelet)
    if not isinstance(coeffs[0], jnp.ndarray):
        raise ValueError(
            "First element of coeffs must be the approximation coefficient tensor."
        )
    _, _, rec_lo, rec_hi = _get_filter_arrays(wavelet, flip=True, dtype=coeffs[0].dtype)
    filt_len = rec_lo.shape[-1]
    rec_filt = _construct_2d_filt(lo=rec_lo, hi=rec_hi)
    rec_filt = jnp.transpose(rec_filt, [1, 0, 2, 3])

    res_ll = jnp.expand_dims(coeffs[0], 1)
    for c_pos, res_lh_hl_hh in enumerate(coeffs[1:]):
        res_ll = jnp.concatenate(
            [
                res_ll,
                jnp.expand_dims(res_lh_hl_hh[0], 1),
                jnp.expand_dims(res_lh_hl_hh[1], 1),
                jnp.expand_dims(res_lh_hl_hh[2], 1),
            ],
            1,
        )
        res_ll = jax.lax.conv_transpose(
            lhs=res_ll,
            rhs=rec_filt,
            padding="VALID",
            strides=[2, 2],
            dimension_numbers=("NCHW", "OIHW", "NCHW"),
            precision=jax.lax.Precision(precision),
        )
        # remove the padding
        padl = (2 * filt_len - 3) // 2
        padr = (2 * filt_len - 3) // 2
        padt = (2 * filt_len - 3) // 2
        padb = (2 * filt_len - 3) // 2
        if c_pos < len(coeffs) - 2:
            pred_len = res_ll.shape[-1] - (padl + padr)
            next_len = coeffs[c_pos + 2][0].shape[-1]
            pred_len2 = res_ll.shape[-2] - (padt + padb)
            next_len2 = coeffs[c_pos + 2][0].shape[-2]
            if next_len != pred_len:
                padr += 1
                pred_len = res_ll.shape[-1] - (padl + padr)
                assert (
                    next_len == pred_len
                ), "padding error, please open an issue on github "
            if next_len2 != pred_len2:
                padb += 1
                pred_len2 = res_ll.shape[-2] - (padt + padb)
                assert (
                    next_len2 == pred_len2
                ), "padding error, please open an issue on github "
        # print('padding', padt, padb, padl, padr)
        if padt > 0:
            res_ll = res_ll[..., padt:, :]
        if padb > 0:
            res_ll = res_ll[..., :-padb, :]
        if padl > 0:
            res_ll = res_ll[..., padl:]
        if padr > 0:
            res_ll = res_ll[..., :-padr]
    return res_ll.squeeze(1)


def _construct_2d_filt(lo: jnp.ndarray, hi: jnp.ndarray) -> jnp.ndarray:
    """Construct 2d filters from 1d inputs using outer products.

    Args:
        lo (jnp.ndarray): 1d lowpass input filter of size [1, length].
        hi (jnp.ndarray): 1d highpass input filter of size [1, length].

    Returns
        jnp.array: 2d filter arrays of shape [4, 1, length, length].
    """
    ll = jnp.outer(lo, lo)
    lh = jnp.outer(hi, lo)
    hl = jnp.outer(lo, hi)
    hh = jnp.outer(hi, hi)
    filt = jnp.stack([ll, lh, hl, hh], 0)
    filt = jnp.expand_dims(filt, 1)
    return filt


def _fwt_pad2d(data: jnp.ndarray, filt_len: int, mode: str = "reflect") -> jnp.ndarray:
    padr = 0
    padl = 0
    padt = 0
    padb = 0
    if filt_len > 2:
        # we pad half of the total requried padding on each side.
        padr += (2 * filt_len - 3) // 2
        padl += (2 * filt_len - 3) // 2
        padt += (2 * filt_len - 3) // 2
        padb += (2 * filt_len - 3) // 2

    # pad to even singal length.
    if data.shape[-1] % 2 != 0:
        padr += 1
    if data.shape[-2] % 2 != 0:
        padb += 1

    data = jnp.pad(data, ((0, 0), (0, 0), (padt, padb), (padl, padr)), mode)
    return data



"""Various utility functions."""

# -*- coding: utf-8 -*-
from collections import namedtuple
from typing import List, Tuple, Union

import jax.numpy as jnp
import pywt

__all__ = ["flatten_2d_coeff_lst"]

Wavelet = namedtuple("Wavelet", ["dec_lo", "dec_hi", "rec_lo", "rec_hi"])


def flatten_2d_coeff_lst(
    coeff_list_2d: List[
        Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]
    ],
    flatten_arrays: bool = True,
) -> List[jnp.ndarray]:
    """Flattens a list of array tuples into a single list.

    Args:
        coeff_list_2d (list): A pywt-style coefficient list.
        flatten_arrays (bool): If true,
             2d array are flattened. Defaults to True.

    Returns:
        list: A single 1-d list with all original elements.
    """
    flat_coeff_lst = []
    for coeff in coeff_list_2d:
        if isinstance(coeff, tuple):
            for c in coeff:
                if flatten_arrays:
                    flat_coeff_lst.append(c.flatten())
                else:
                    flat_coeff_lst.append(c)
        else:
            if flatten_arrays:
                flat_coeff_lst.append(coeff.flatten())
            else:
                flat_coeff_lst.append(coeff)
    return flat_coeff_lst


def _as_wavelet(wavelet: Union[Wavelet, str]) -> pywt.Wavelet:
    """Ensure the input argument to be a pywt wavelet compatible object.

    Args:
        wavelet (Wavelet or str): The input argument, which is either a
            pywt wavelet compatible object or a valid pywt wavelet name string.

    Returns:
        pywt.Wavelet: the input wavelet object or the pywt wavelet object
            described by the input str.
    """
    if isinstance(wavelet, str):
        return pywt.Wavelet(wavelet)
    else:
        return wavelet



"""Runs the example from https://github.com/google/jax#automatic-differentiation-with-grad ."""

#
# Created on Thu Jun 11 2020
# Copyright (c) 2020 Moritz Wolter
#


import jax.numpy as np
from jax import grad


def test_jax():
    """Test if jax is correctly installed."""

    def tanh(x):  # Define a function
        y = np.exp(-2.0 * x)
        return (1.0 - y) / (1.0 + y)

    grad_tanh = grad(tanh)  # Obtain its gradient function
    print(grad_tanh(1.0))  # Evaluate it at x = 1.0
    assert (grad_tanh(1.0) - 0.4199743) < 0.000001



"""Convolution fast wavelet transform test code."""

#
# Created on Thu Jun 11 2020
# Copyright (c) 2020 Moritz Wolter
#

import jax.numpy as jnp
import numpy as np
import pytest
import pywt
from jax.config import config

from src.jaxwt.conv_fwt import wavedec, waverec
from tests._lorenz import generate_lorenz

config.update("jax_enable_x64", True)


def test_haar_fwt_ifwt_16_float32():
    """Test Haar wavelet analysis and synthesis on 16 sample signal."""
    wavelet = pywt.Wavelet("haar")
    data = jnp.array(
        [
            1.0,
            2.0,
            3.0,
            4.0,
            5.0,
            6.0,
            7.0,
            8.0,
            9.0,
            10.0,
            11.0,
            12.0,
            13.0,
            14.0,
            15.0,
            16.0,
        ]
    ).astype(jnp.float32)
    data = jnp.expand_dims(data, 0)
    coeffs_pywt = pywt.wavedec(data, wavelet, level=2)
    coeffs_jaxwt = wavedec(data, wavelet, level=2)
    cat_coeffs_pywt = jnp.concatenate(coeffs_pywt, -1)
    cat_coeffs_jaxwt = jnp.concatenate(coeffs_jaxwt, -1)
    assert jnp.allclose(cat_coeffs_pywt, cat_coeffs_jaxwt)
    reconstructed_data = waverec(coeffs_jaxwt, wavelet)
    assert jnp.allclose(reconstructed_data, data)


@pytest.mark.parametrize("wavelet", ["haar", "db2", "db3", "sym4"])
@pytest.mark.parametrize("mode", ["reflect", "symmetric", "zero"])
@pytest.mark.parametrize("tmax", [1.27, 1.26])
@pytest.mark.parametrize("level", [1, 2, None])
def test_fwt_ifwt_lorenz(wavelet, level, mode, tmax):
    """Test wavelet analysis and synthesis on lorenz signal."""
    wavelet = pywt.Wavelet(wavelet)
    lorenz = jnp.transpose(
        jnp.expand_dims(generate_lorenz(tmax=tmax)[:, 0], -1), [1, 0]
    ).astype(jnp.float64)
    coeff = wavedec(lorenz, wavelet, mode=mode, level=level)
    pywt_coeff = pywt.wavedec(lorenz, wavelet, mode=mode, level=level)
    jwt_cat_coeff = jnp.concatenate(coeff, axis=-1).squeeze()
    pywt_cat_coeff = jnp.concatenate(pywt_coeff, axis=-1).squeeze()
    assert jnp.allclose(jwt_cat_coeff, pywt_cat_coeff)
    rec_data = waverec(coeff, wavelet)
    assert jnp.allclose(rec_data[..., : lorenz.shape[-1]], lorenz)


@pytest.mark.parametrize("wavelet", ["db2", "sym4"])
@pytest.mark.parametrize("mode", ["reflect", "zero"])
@pytest.mark.parametrize("batch_size", [1, 3])
@pytest.mark.parametrize("level", [2, None])
@pytest.mark.parametrize("dtype", [jnp.float64, jnp.float32])
def test_batch_fwt_ifwt(wavelet, mode, batch_size, level, dtype: jnp.dtype):
    """Test the batched version of the fwt."""
    if dtype == jnp.float32:
        atol = 1e-4
    else:
        atol = 1e-8

    wavelet = pywt.Wavelet(wavelet)
    random_dat = jnp.array(np.random.randn(batch_size, 100)).astype(dtype)
    coeff = wavedec(random_dat, wavelet, mode=mode, level=level)
    rec_data = waverec(coeff, wavelet)
    assert jnp.allclose(rec_data[..., : random_dat.shape[-1]], random_dat, atol=atol)



"""Test jit compilation."""
from collections import namedtuple

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import pywt

import src.jaxwt as jaxwt
from tests._lorenz import generate_lorenz

WaveletTuple = namedtuple("Wavelet", ["dec_lo", "dec_hi", "rec_lo", "rec_hi"])


def _to_wavelet_tuple(wavelet: pywt.Wavelet) -> WaveletTuple:
    return WaveletTuple(
        jnp.array(wavelet.dec_lo),
        jnp.array(wavelet.dec_hi),
        jnp.array(wavelet.rec_lo),
        jnp.array(wavelet.rec_hi),
    )


@pytest.mark.slow
@pytest.mark.parametrize("wavelet_string", ["db1", "db4", "sym5"])
@pytest.mark.parametrize("level", [1, 2])
@pytest.mark.parametrize("batch_size", [1, 3])
@pytest.mark.parametrize("length", [64, 65])
@pytest.mark.parametrize("dtype", [jnp.float32, jnp.float64])
def test_conv_fwt_jit(wavelet_string, level, length, batch_size, dtype):
    """Test jitting a convolution fwt, for various levels and padding options."""
    data = generate_lorenz().transpose()[:batch_size, :length].astype(dtype)

    wavelet = pywt.Wavelet(wavelet_string)
    # pywt.Wavelets do not compile with jax.jit
    wavelet = _to_wavelet_tuple(wavelet)
    jit_wavedec = jax.jit(jaxwt.wavedec, static_argnames=["level"])
    coeff = jit_wavedec(data, wavelet, level=level)
    jit_waverec = jax.jit(jaxwt.waverec)
    res = jit_waverec(coeff, wavelet)
    assert jnp.allclose(data, res[:, : data.shape[-1]])


@pytest.mark.parametrize("level", [1, 2])
def test_conv_fwt_jit_2d(level):
    """Test the jit compilation feature for the wavedec2 function."""
    data = jnp.array(np.random.randn(10, 64, 64)).astype(jnp.float64)
    wavelet = pywt.Wavelet("db2")
    wavelet = _to_wavelet_tuple(wavelet)
    jit_wavedec2 = jax.jit(jaxwt.wavedec2, static_argnames=["level"])
    coeff = jit_wavedec2(data, wavelet, level=level)
    jit_waverec2 = jax.jit(jaxwt.waverec2)
    rec = jit_waverec2(coeff, wavelet=wavelet)
    assert np.allclose(rec, data)



"""Compute wavelet packets using jwt."""

#
# Created on Fri Jun 19 2020
# Copyright (c) 2020 Moritz Wolter
#
import collections
from itertools import product
from typing import TYPE_CHECKING, List, Optional, Union

# import jax
import jax.numpy as jnp
import pywt

from .conv_fwt import wavedec, waverec
from .conv_fwt_2d import wavedec2, waverec2
from .utils import _as_wavelet

if TYPE_CHECKING:
    BaseDict = collections.UserDict[str, jnp.ndarray]
else:
    BaseDict = collections.UserDict


class WaveletPacket(BaseDict):
    """A wavelet packet tree."""

    def __init__(
        self,
        data: jnp.ndarray,
        wavelet: pywt.Wavelet,
        mode: str = "reflect",
        max_level: Optional[int] = None,
    ):
        """Create a wavelet packet decomposition object.

        Args:
            data (jnp.ndarray): The input data array of shape [batch_size, time].
            wavelet (pywt.Wavelet): The wavelet used for the decomposition.
            mode (str): The desired padding method. Choose i.e.
                "reflect", "symmetric" or "zero". Defaults to "reflect".

        Example:
            >>> import pywt
            >>> import jax.numpy as jnp
            >>> from jaxwt import WaveletPacket
            >>> import scipy.signal as signal
            >>> wavelet = pywt.Wavelet("db4")
            >>> t = jnp.linspace(0, 10, 5001)
            >>> w = signal.chirp(t, f0=0.00001,
            >>>                  f1=20, t1=10, method="linear")
            >>> wp = WaveletPacket(data=w, wavelet=wavelet)
            >>> nodes = wp.get_level(7)
            >>> jnp_lst = []
            >>> for node in nodes:
            >>>     jnp_lst.append(wp[node])
            >>> viz = jnp.concatenate(jnp_lst)
        """
        if len(data.shape) == 1:
            self.input_data = jnp.expand_dims(data, 0)
        elif len(data.shape) == 2:
            self.input_data = data

        self.wavelet = _as_wavelet(wavelet)
        self.mode = mode
        self.data = {}
        if max_level is None:
            self.max_level = pywt.dwt_max_level(data.shape[-1], self.wavelet.dec_len)
        else:
            self.max_level = max_level
        self._recursive_dwt(self.input_data, level=0, path="")

    def get_level(self, level: int) -> List[str]:
        """Return the graycodes for a given level.

        Args:
            level (int): The required depth of the tree.

        Returns:
            list: A list with the node names.
        """
        return self._get_graycode_order(level)

    def _get_graycode_order(self, level: int, x: str = "a", y: str = "d") -> List[str]:
        graycode_order = [x, y]
        for _ in range(level - 1):
            graycode_order = [x + path for path in graycode_order] + [
                y + path for path in graycode_order[::-1]
            ]
        if level == 0:
            return [""]
        else:
            return graycode_order

    def _recursive_dwt(
        self,
        data: jnp.ndarray,
        level: int,
        path: str,
    ) -> None:
        self.data[path] = data
        if level < self.max_level:
            res_lo, res_hi = wavedec(data, self.wavelet, 1, mode=self.mode)
            self._recursive_dwt(res_lo, level + 1, path + "a")
            self._recursive_dwt(res_hi, level + 1, path + "d")
        else:
            self.data[path] = data

    def reconstruct(self) -> "WaveletPacket":
        """Recursively reconstruct the input starting from the leaf nodes.

        Reconstruction replaces the input-data originally assigned to this object.

        Note:
           Only changes to leaf node data impacts the results,
           since changes in all other nodes will be replaced with
           a reconstruction from the leafs.

        Example:
            >>> import jaxwt as jwt
            >>> import jax
            >>> key = jax.random.PRNGKey(0)
            >>> input_data = jax.random.normal(key, (1, 24))
            >>> jwp = jwt.WaveletPacket(input_data, "haar", max_level=2)
            >>> jwp["a" * 2] *= 0
            >>> jwp.reconstruct()
            >>> print(jwp[""])
        """
        if self.max_level is None:
            self.max_level = pywt.dwt_max_level(
                self[""].shape[-1], self.wavelet.dec_len
            )

        for level in reversed(range(self.max_level)):
            for node in self.get_level(level):
                data_a = self[node + "a"]
                data_b = self[node + "d"]
                rec = waverec([data_a, data_b], self.wavelet)
                self[node] = rec
        return self


class WaveletPacket2D(BaseDict):
    """A wavelet packet tree."""

    def __init__(
        self,
        data: jnp.ndarray,
        wavelet: Union[str, pywt.Wavelet],
        mode: str = "reflect",
        max_level: Optional[int] = None,
    ):
        """Create a 2D-wavelet packet decomposition object.

        Example code illustrating the use of this class is available at:
        https://github.com/v0lta/Jax-Wavelet-Toolbox/tree/packet-patch/examples/deepfake_analysis

        Args:
            data (jnp.ndarray): The input data array of shape [batch_size, height, width].
            wavelet (pywt.Wavelet or str): The wavelet used for the decomposition.
            mode (str): The desired padding method. Choose i.e.
                "reflect", "symmetric" or "zero". Defaults to "reflect".
            max_level (int, optional): Choose the desired decomposition level.
        """
        self.input_data = data
        self.wavelet: pywt.Wavelet = _as_wavelet(wavelet)
        self.mode = mode
        self.data = {}
        if max_level is None:
            self.max_level = pywt.dwt_max_level(
                min(self.input_data.shape[-2:]), self.wavelet.dec_len
            )
        else:
            self.max_level = max_level
        self._recursive_dwt2d(self.input_data, level=0, path="")

    def _recursive_dwt2d(
        self,
        data: jnp.ndarray,
        level: int,
        path: str,
    ) -> None:
        self.data[path] = data
        if level < self.max_level:
            result_a, (result_h, result_v, result_d) = wavedec2(
                data, self.wavelet, 1, mode=self.mode
            )
            # assert for type checking
            assert not isinstance(result_a, tuple)
            self._recursive_dwt2d(result_a, level + 1, path + "a")
            self._recursive_dwt2d(result_h, level + 1, path + "h")
            self._recursive_dwt2d(result_v, level + 1, path + "v")
            self._recursive_dwt2d(result_d, level + 1, path + "d")
        else:
            self.data[path] = data

    def _get_natural_order(self, level: int) -> List[str]:
        """Get the natural ordering for a given decomposition level.

        Args:
            level (int): The decomposition level.

        Returns:
            list: A list with the filter order strings.
        """
        return ["".join(p) for p in list(product(["a", "h", "v", "d"], repeat=level))]

    def reconstruct(self) -> "WaveletPacket2D":
        """Recursively reconstruct the input starting from the leaf nodes.

        Note:
           Only changes to leaf node data impacts the results,
           since changes in all other nodes will be replaced with
           a reconstruction from the leafs.
        """
        if self.max_level is None:
            self.max_level = pywt.dwt_max_level(
                min(self[""].shape[-2:]), self.wavelet.dec_len
            )

        for level in reversed(range(self.max_level)):
            for node in self._get_natural_order(level):
                data_a = self[node + "a"]
                data_h = self[node + "h"]
                data_v = self[node + "v"]
                data_d = self[node + "d"]
                rec = waverec2([data_a, (data_h, data_v, data_d)], self.wavelet)
                self[node] = rec
        return self



"""Convolution based fast wavelet transforms."""

# -*- coding: utf-8 -*-

# Created on Thu Jun 11 2020
# Copyright (c) 2020 Moritz Wolter
#
from typing import List, Optional, Tuple, Union

import jax
import jax.lax
import jax.numpy as jnp
import pywt

from .utils import _as_wavelet


def wavedec(
    data: jnp.ndarray,
    wavelet: pywt.Wavelet,
    level: Optional[int] = None,
    mode: str = "reflect",
    precision: str = "highest",
) -> List[jnp.ndarray]:
    """Compute the analysis wavelet transform of the last dimension.

    Args:
        data (jnp.ndarray): Input data array of shape [batch, time].
        wavelet (pywt.Wavelet): The named tuple containing the wavelet
                    filter arrays.
        level (int): Max scale level to be used,
                     of none as many levels as possible are
                     used. Defaults to None.
        mode (str): The padding used to extend the input signal. Choose reflect, symmetric or zero.
            Defaults to reflect.
        precision (str): The desired precision, choose "fastest", "high" or "highest".
            Defaults to "highest".


    Returns:
        list: List containing the wavelet coefficients of shape [batch_size, coefficients].
            The coefficients are in pywt order:
            [cA_n, cD_n, cD_n-1, …, cD2, cD1].
            A denotes approximation and D detail coefficients.

    Raises:
        ValueError: If the dimensionality of the input data array is unsupported.

    Examples:
        >>> import pywt
        >>> import jaxwt as jwt
        >>> import jax.numpy as jnp
        >>> # generate an input of even length.
        >>> data = jnp.array([0., 1., 2., 3, 4, 5, 5, 4, 3, 2, 1, 0])
        >>> jwt.wavedec(data, wavelet=pywt.Wavelet('haar'), level=2)
    """
    wavelet = _as_wavelet(wavelet)
    if len(data.shape) == 1:
        # add channel and batch dimension.
        data = jnp.expand_dims(jnp.expand_dims(data, 0), 0)
    elif len(data.shape) == 2:
        # add the channel dimension.
        data = jnp.expand_dims(data, 1)
    else:
        raise ValueError(
            "Wavedec only supports up to two input dimensions. \
             Optionally-batched one dimensional inputs work."
        )

    dec_lo, dec_hi, _, _ = _get_filter_arrays(wavelet, flip=True, dtype=data.dtype)
    filt_len = dec_lo.shape[-1]
    filt = jnp.stack([dec_lo, dec_hi], 0)

    if level is None:
        level = pywt.dwt_max_level(data.shape[-1], filt_len)

    result_lst = []
    res_lo = data
    for _ in range(level):
        res_lo = _fwt_pad(res_lo, len(wavelet.dec_lo), mode=mode)
        res = jax.lax.conv_general_dilated(
            lhs=res_lo,  # lhs = NCH image tensor
            rhs=filt,  # rhs = OIH conv kernel tensor
            padding="VALID",
            window_strides=[
                2,
            ],
            dimension_numbers=("NCH", "OIH", "NCH"),
            precision=jax.lax.Precision(precision),
        )
        res_lo, res_hi = jnp.split(res, 2, 1)
        result_lst.append(res_hi.squeeze(1))
    result_lst.append(res_lo.squeeze(1))
    result_lst.reverse()
    return result_lst


def waverec(
    coeffs: List[jnp.ndarray],
    wavelet: pywt.Wavelet,
    precision: str = "highest",
) -> jnp.ndarray:
    """Reconstruct the original signal in one dimension.

    Args:
        coeffs (list): Wavelet coefficients, typically produced by the wavedec function.
            List entries of shape [batch_size, coefficients] work.
        wavelet (pywt.Wavelet): The named tuple containing the wavelet filters used to evaluate
                              the decomposition.
        precision (str): The desired precision, choose "fastest", "high" or "highest".
            Defaults to "highest".

    Returns:
        jnp.array: Reconstruction of the original data.

    Examples:
        >>> import pywt
        >>> import jaxwt as jwt
        >>> import jax.numpy as jnp
        >>> # generate an input of even length.
        >>> data = jnp.array([0., 1., 2., 3, 4, 5, 5, 4, 3, 2, 1, 0])
        >>> transformed = jwt.wavedec(data, pywt.Wavelet('haar'))
        >>> jwt.waverec(transformed, pywt.Wavelet('haar'))
    """
    wavelet = _as_wavelet(wavelet)
    # lax's transpose conv requires filter flips in contrast to pytorch.
    _, _, rec_lo, rec_hi = _get_filter_arrays(wavelet, flip=True, dtype=coeffs[0].dtype)
    filt_len = rec_lo.shape[-1]
    filt = jnp.stack([rec_lo, rec_hi], 1)

    res_lo = coeffs[0]
    for c_pos, res_hi in enumerate(coeffs[1:]):
        # print('shapes', res_lo.shape, res_hi.shape)
        res_lo = jnp.stack([res_lo, res_hi], 1)
        res_lo = jax.lax.conv_transpose(
            lhs=res_lo,
            rhs=filt,
            padding="VALID",
            strides=[
                2,
            ],
            dimension_numbers=("NCH", "OIH", "NCH"),
            precision=jax.lax.Precision(precision),
        )
        res_lo = _fwt_unpad(res_lo, filt_len, c_pos, coeffs)
        res_lo = res_lo.squeeze(1)
    return res_lo


def _fwt_unpad(
    res_lo: jnp.ndarray, filt_len: int, c_pos: int, coeffs: List[jnp.ndarray]
) -> jnp.ndarray:
    padr = 0
    padl = 0
    if filt_len > 2:
        padr += (2 * filt_len - 3) // 2
        padl += (2 * filt_len - 3) // 2
    if c_pos < len(coeffs) - 2:
        pred_len = res_lo.shape[-1] - (padl + padr)
        nex_len = coeffs[c_pos + 2].shape[-1]
        if nex_len != pred_len:
            padl += 1
            pred_len = res_lo.shape[-1] - padl
    if padl == 0:
        res_lo = res_lo[..., padr:]
    else:
        res_lo = res_lo[..., padr:-padl]
    return res_lo


def _fwt_pad(data: jnp.ndarray, filt_len: int, mode: str = "reflect") -> jnp.ndarray:
    """Pad an input to ensure our fwts are invertible.

    Args:
        data (jnp.ndarray): The input array.
        filt_len (int): The length of the wavelet filters
        mode (str): How to pad. Defaults to "reflect".

    Returns:
        jnp.array: A padded version of the input data array.
    """
    # pad to we see all filter positions and pywt compatability.
    # convolution output length:
    # see https://arxiv.org/pdf/1603.07285.pdf section 2.3:
    # floor([data_len - filt_len]/2) + 1
    # should equal pywt output length
    # floor((data_len + filt_len - 1)/2)
    # => floor([data_len + total_pad - filt_len]/2) + 1
    #    = floor((data_len + filt_len - 1)/2)
    # (data_len + total_pad - filt_len) + 2 = data_len + filt_len - 1
    # total_pad = 2*filt_len - 3

    if mode == "zero":
        # translate pywt to numpy.
        mode = "constant"

    padr = (2 * filt_len - 3) // 2
    padl = (2 * filt_len - 3) // 2

    # pad to even singal length.
    if data.shape[-1] % 2 != 0:
        padr += 1

    data = jnp.pad(data, ((0, 0), (0, 0), (padl, padr)), mode)
    return data


def _get_filter_arrays(
    wavelet: pywt.Wavelet, flip: bool, dtype: jnp.dtype = jnp.float64  # type: ignore
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Extract the filter coefficients from an input wavelet object.

    Args:
        wavelet (pywt.Wavelet): A pywt-style input wavelet.
        flip (bool): If true flip the input coefficients.
        dtype: The desired precision. Defaults to jnp.float64 .

    Returns:
        tuple: The dec_lo, dec_hi, rec_lo and rec_hi
            filter coefficients as jax arrays.
    """

    def create_array(filter: Union[List[float], jnp.ndarray]) -> jnp.ndarray:
        if flip:
            if type(filter) is jnp.ndarray:
                return jnp.expand_dims(jnp.flip(filter), 0)
            else:
                return jnp.expand_dims(jnp.array(filter[::-1]), 0)
        else:
            if type(filter) is jnp.ndarray:
                return jnp.expand_dims(filter, 0)
            else:
                return jnp.expand_dims(jnp.array(filter), 0)

    if isinstance(wavelet, str):
        wavelet = pywt.Wavelet(wavelet)
        dec_lo, dec_hi, rec_lo, rec_hi = wavelet.filter_bank
    elif type(wavelet) is pywt.Wavelet:
        dec_lo, dec_hi, rec_lo, rec_hi = wavelet.filter_bank
    else:
        dec_lo, dec_hi, rec_lo, rec_hi = wavelet
    dec_lo = create_array(dec_lo).astype(dtype)
    dec_hi = create_array(dec_hi).astype(dtype)
    rec_lo = create_array(rec_lo).astype(dtype)
    rec_hi = create_array(rec_hi).astype(dtype)
    return dec_lo, dec_hi, rec_lo, rec_hi



"""Jax compatible cwt code."""
from typing import Tuple, Union

import jax.numpy as jnp
import numpy as np
from jax.numpy.fft import fft, ifft
from pywt import ContinuousWavelet, DiscreteContinuousWavelet, Wavelet
from pywt._functions import integrate_wavelet, scale2frequency


def _next_fast_len(n: int) -> int:
    """Round up size to the nearest power of two.

    Given a number of samples `n`, returns the next power of two
    following this number to take advantage of FFT speedup.
    This fallback is less efficient than `scipy.fftpack.next_fast_len`
    """
    return int(2 ** jnp.ceil(jnp.log2(n)))


def cwt(
    data: jnp.ndarray,
    scales: Union[np.ndarray, jnp.ndarray],  # type: ignore
    wavelet: Union[ContinuousWavelet, str],
    sampling_period: float = 1.0,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the single dimensional continuous wavelet transform.

    This function is a jax port of pywt.cwt as found at:
    https://github.com/PyWavelets/pywt/blob/master/pywt/_cwt.py

    Args:
        data (jnp.ndarray): The ijnput tensor of shape [batch_size, time].
        scales (np.ndarray or jnp.array):
            The wavelet scales to use. One can use
            ``f = pywt.scale2frequency(wavelet, scale)/sampling_period`` to determine
            what physical frequency, ``f``. Here, ``f`` is in hertz when the
            ``sampling_period`` is given in seconds.
            wavelet (str or Wavelet of ContinuousWavelet): The wavelet to work with.
        wavelet (ContinuousWavelet or str): The continuous wavelet to work with.
        sampling_period (float): Sampling period for the frequencies output (optional).
            The values computed for ``coefs`` are independent of the choice of
            ``sampling_period`` (i.e. ``scales`` is not scaled by the sampling
            period).

    Raises:
        ValueError: If a scale is too small for the input signal.

    Returns:
        Tuple[jnp.ndarray, jnp.ndarray]: A tuple with the transformation matrix
            and frequencies in this order.

    Example:
        >>> import jaxwt as jwt
        >>> import jax.numpy as jnp
        >>> import scipy.signal as signal
        >>> t = jnp.linspace(-2, 2, 800, endpoint=False)
        >>> sig = signal.chirp(t, f0=1, f1=12, t1=2, method="linear")
        >>> widths = jnp.arange(1, 31)
        >>> cwtmatr, freqs = jwt.cwt(
        >>>     jnp.array(sig), widths, "mexh",
        >>>     sampling_period=(4 / 800) * jnp.pi
        >>> )
    """
    # accept array_like ijnput; make a copy to ensure a contiguous array
    if not isinstance(wavelet, (ContinuousWavelet, Wavelet)):
        wavelet = DiscreteContinuousWavelet(wavelet)
    if type(scales) is jnp.ndarray:
        scales = np.array(scales)
    elif jnp.isscalar(scales):
        scales = jnp.array([scales])
    # if not jnp.isscalar(axis):
    #    raise jnp.AxisError("axis must be a scalar.")

    precision = 10
    int_psi, x = integrate_wavelet(wavelet, precision=precision)
    if type(wavelet) is ContinuousWavelet:
        int_psi = jnp.conj(int_psi) if wavelet.complex_cwt else int_psi
    int_psi = jnp.array(int_psi)

    # convert int_psi, x to the same precision as the data
    # x = jnp.asarray(x, dtype=data.cpu().numpy().real.dtype)
    x = jnp.array(x)

    size_scale0 = -1
    fft_data = None

    out = []
    for scale in scales:
        step = x[1] - x[0]
        j = jnp.arange(scale * (x[-1] - x[0]) + 1) / (scale * step)
        j = j.astype(int)  # floor
        if j[-1] >= len(int_psi):
            j = jnp.extract(j < len(int_psi), j)
        int_psi_scale = jnp.flip(int_psi[j], 0)

        # The padding is selected for:
        # - optimal FFT complexity
        # - to be larger than the two signals length to avoid circular
        #   convolution
        size_scale = _next_fast_len(data.shape[-1] + len(int_psi_scale) - 1)
        if size_scale != size_scale0:
            # Must recompute fft_data when the padding size changes.
            fft_data = fft(data, size_scale, axis=-1)
        size_scale0 = size_scale
        fft_wav = fft(int_psi_scale, size_scale, axis=-1)
        conv = ifft(fft_wav * fft_data, axis=-1)
        conv = conv[..., : data.shape[-1] + len(int_psi_scale) - 1]

        coef = -jnp.sqrt(scale) * jnp.diff(conv, axis=-1)

        # transform axis is always -1
        d = (coef.shape[-1] - data.shape[-1]) / 2.0
        if d > 0:
            coef = coef[..., int(jnp.floor(d)) : -int(jnp.ceil(d))]
        elif d < 0:
            raise ValueError("Selected scale of {} too small.".format(scale))

        out.append(coef)
    out_tensor = jnp.stack(out)
    if type(wavelet) is Wavelet:
        out_tensor = out_tensor.real
    else:
        out_tensor = out_tensor if wavelet.complex_cwt else out_tensor.real

    frequencies = scale2frequency(wavelet, scales, precision)
    if jnp.isscalar(frequencies):
        frequencies = jnp.array([frequencies])
    frequencies /= sampling_period
    return out_tensor, frequencies



# -*- coding: utf-8 -*-

"""Differentiable and gpu enabled fast wavelet transforms in JAX."""
from .continuous_transform import cwt
from .conv_fwt import wavedec, waverec
from .conv_fwt_2d import wavedec2, waverec2
from .packets import WaveletPacket, WaveletPacket2D



"""2d Convolution fast wavelet transform test code."""
import jax.numpy as jnp
import pytest
import pywt
import scipy.datasets
from jax.config import config

from src.jaxwt.conv_fwt_2d import wavedec2, waverec2
from src.jaxwt.utils import flatten_2d_coeff_lst

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")


@pytest.mark.parametrize("mode", ["symmetric", "reflect"])
@pytest.mark.parametrize("wavelet", ["haar", "db2", "db3", "sym4"])
@pytest.mark.parametrize("level", [1, 2, None])
@pytest.mark.parametrize("size", [(65, 65), (64, 64), (47, 45), (45, 47)])
@pytest.mark.parametrize("dtype", [jnp.float64, jnp.float32])
def test_conv_2d(wavelet: str, level: int, size: tuple, mode: str, dtype: jnp.dtype):
    """Run a specific test."""
    if dtype == jnp.float32:
        atol = 1e-3
    else:
        atol = 1e-8

    wavelet = pywt.Wavelet(wavelet)
    face = jnp.transpose(scipy.datasets.face(), [2, 0, 1]).astype(dtype)
    face = face[:, 128 : (128 + size[0]), 256 : (256 + size[1])]

    coeff2d_pywt = pywt.wavedec2(face, wavelet, mode=mode, level=level)
    coeff2d = wavedec2(face, wavelet, level=level, mode=mode)
    # test pywt compatability
    pywt_flat_list = jnp.concatenate(flatten_2d_coeff_lst(coeff2d_pywt), -1)
    jwt_flat_list = jnp.concatenate(flatten_2d_coeff_lst(coeff2d), -1)
    assert jnp.allclose(pywt_flat_list, jwt_flat_list, atol=atol)

    # test invertability
    reconstruction_2d = waverec2(coeff2d, wavelet)[..., : size[0], : size[1]]
    assert jnp.allclose(reconstruction_2d, face, atol=atol)



"""Test the continuous transformation code."""
from typing import Union

import jax.numpy as jnp
import numpy as np
import pytest
import pywt
from scipy import signal

from src.jaxwt.continuous_transform import cwt

continuous_wavelets = [
    "cgau1",
    "cgau2",
    "cgau3",
    "cgau4",
    "cgau5",
    "cgau6",
    "cgau7",
    "cgau8",
    "gaus1",
    "gaus2",
    "gaus3",
    "gaus4",
    "gaus5",
    "gaus6",
    "gaus7",
    "gaus8",
    "mexh",
    "morl",
]


@pytest.mark.parametrize("scales", [np.arange(1, 16), 5.0, np.arange(1, 15)])
@pytest.mark.parametrize("samples", [31, 32])
@pytest.mark.parametrize("wavelet", continuous_wavelets)
def test_cwt(
    wavelet: str, samples: int, scales: Union[np.ndarray, np.ndarray, float]
) -> None:
    """Test the cwt implementation for various wavelets."""
    t = np.linspace(-1, 1, samples, endpoint=False)
    sig = signal.chirp(t, f0=1, f1=50, t1=10, method="linear")
    cwtmatr, freqs = pywt.cwt(data=sig, scales=scales, wavelet=wavelet)
    cwtmatr_jax, freqs_jax = cwt(jnp.array(sig), jnp.array(scales), wavelet)
    assert jnp.allclose(cwtmatr_jax, cwtmatr)
    assert jnp.allclose(freqs, freqs_jax)


@pytest.mark.parametrize("wavelet", continuous_wavelets)
def test_cwt_batched(wavelet):
    """Test batched transforms."""
    sig = np.random.randn(10, 200)
    widths = np.arange(1, 30)
    cwtmatr, freqs = pywt.cwt(sig, widths, wavelet)
    cwtmatr_jax, freqs_jax = cwt(jnp.array(sig), widths, wavelet)
    assert jnp.allclose(cwtmatr_jax, cwtmatr)
    assert jnp.allclose(freqs, freqs_jax)



"""Test the wavelet packet code."""
#
# Created on Fri Jun 19 2020
# Copyright (c) 2020 Moritz Wolter
#

from itertools import product

import jax.numpy as jnp
import numpy as np
import pytest
import pywt
from jax.config import config

from src.jaxwt.packets import WaveletPacket, WaveletPacket2D

config.update("jax_enable_x64", True)


# TODO: add more input shapes. and level none
@pytest.mark.parametrize("input_shape", ((2, 256), (3, 255), (1, 244)))
@pytest.mark.parametrize("wavelet", ("haar", "db2", "db3"))
@pytest.mark.parametrize("level", (2, 4))
@pytest.mark.parametrize("mode", ("reflect", "symmetric", "zero"))
def test_packets_1d(input_shape, wavelet, level, mode):
    """Test the wavelet packet analysis code."""
    wavelet = pywt.Wavelet(wavelet)
    input_data = jnp.array(np.random.randn(*input_shape))
    jwp = WaveletPacket(input_data, wavelet, mode=mode)
    nodes = jwp.get_level(level)
    jnp_lst = []
    for node in nodes:
        jnp_lst.append(jwp[node])
    jres = jnp.stack(jnp_lst)

    np_input_data = np.array(input_data._value)
    wp = pywt.WaveletPacket(np_input_data, wavelet=wavelet, mode=mode)
    nodes = [node.path for node in wp.get_level(level, "freq")]
    np_lst = []
    for node in nodes:
        np_lst.append(wp[node].data)
    res = jnp.stack(np_lst)
    assert jnp.allclose(jres, res)


@pytest.mark.parametrize("input_shape", ((1, 32), (3, 255), (1, 244)))
@pytest.mark.parametrize("base_key", ["a", "d"])
@pytest.mark.parametrize("wavelet", ("haar", "db2", "db3"))
@pytest.mark.parametrize("level", (1, 2, 4))
@pytest.mark.parametrize("mode", ("reflect", "zero"))
def test_inverse_packets_1d(input_shape, wavelet, level, mode, base_key):
    """Test 1d packet inversion."""
    wavelet = pywt.Wavelet(wavelet)
    input_data = jnp.array(np.random.randn(*input_shape))
    jwp = WaveletPacket(input_data, wavelet, mode=mode, max_level=level)
    np_input_data = np.array(input_data._value)
    wp = pywt.WaveletPacket(np_input_data, wavelet=wavelet, mode=mode, maxlevel=level)

    wp[base_key * level].data *= 0
    jwp[base_key * level] *= 0

    wp.reconstruct(update=True)
    jwp.reconstruct()

    assert jnp.allclose(wp[""].data, jwp[""][:, : input_shape[-1]])


@pytest.mark.parametrize("input_shape", ((2, 32, 32), (3, 33, 33), (1, 32, 33)))
@pytest.mark.parametrize("wavelet", ("haar", "db2"))
@pytest.mark.parametrize("level", (2, 3))
@pytest.mark.parametrize("mode", ("reflect", "symmetric", "zero"))
def test_packets_2d(input_shape, wavelet, level, mode):
    """Test the wavelet packet analysis code."""
    wavelet = pywt.Wavelet(wavelet)
    input_data = jnp.array(np.random.randn(*input_shape))
    jwp = WaveletPacket2D(input_data, wavelet, mode=mode)
    jnp_list = []
    wp_keys = list(product(["a", "h", "v", "d"], repeat=level))
    for node in wp_keys:
        jnp_list.append(jwp["".join(node)])
    jres = jnp.stack(jnp_list)

    np_input_data = np.array(input_data._value)
    wp = pywt.WaveletPacket2D(np_input_data, wavelet=wavelet, mode=mode)
    np_lst = []
    for node in wp_keys:
        np_lst.append(wp["".join(node)].data)
    res = jnp.stack(np_lst)
    assert jnp.allclose(jres, res)


@pytest.mark.parametrize("level", [1, 3])
@pytest.mark.parametrize("base_key", ["a", "h", "d"])
@pytest.mark.parametrize("size", [(1, 32, 32), (2, 31, 64)])
@pytest.mark.parametrize("wavelet", ["db1", "db2", "sym4"])
def test_inverse_packet_2d(level, base_key, size, wavelet):
    """Test the 2d reconstruction code."""
    signal = np.random.randn(size[0], size[1], size[2])
    mode = "reflect"
    wp = pywt.WaveletPacket2D(signal, wavelet, mode=mode, maxlevel=level)
    jaxwp = WaveletPacket2D(jnp.array(signal), wavelet, mode=mode, max_level=level)
    wp[base_key * level].data *= 0
    jaxwp[base_key * level] *= 0
    wp.reconstruct(update=True)
    jaxwp.reconstruct()
    assert jnp.allclose(wp[""].data, jaxwp[""][:, : size[1], : size[2]])
